/*
  This code is released under the Antivirial License.

  http://www.ecstaticlyrics.com/antiviral/

  The only condition of the license is that you may not change the license
  terms.  Because of this, you may *not* include this code in GPL-licensed
  projects, as the GPL requires that the entire project be GPL-licensed.
*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>
#include <arpa/inet.h>

#include "memory.h"
#include "glfw.h"
#include "thread.h"
#include "radio.h"
#include "gui.h"
#include "easy.h"
#include "option.h"
#include "common.h"
#include "color.h"

struct waterfall {
  char *buffer;
  int *texture;
  int width;
  int height;
  volatile int index;
  int initialized;
  MUTEX mutex;
};

static int waterfall_pause = 0;

static struct waterfall spectrum = {NULL, NULL, 1024, 768, 0, 0};
static struct waterfall audio = {NULL, NULL, 256, 256, 0, 0};
static struct waterfall signal = {NULL, NULL, 256, 128, 0, 0};

static char iq_sample[2 * 1024];
void user_iq_sample (char *values) {
  memmove(iq_sample, values, 2 * 1024);
};

static void db_to_rgb (float *r, float *g, float *b, float db) {

    // black -> blue  -75 to -60
    // blue -> green  -60 to -45
    // green -> yellow  -45 to -30
    // yellow -> orange  -30 to -15
    // orange -> red  -15 to 0

    if (db < -60.0f) {
      float v = (db - -75.0f) / 15.0f;
      *r = 0.0f;
      *g = 0.25f * v;
      *b = 0.5f * v;
    } else if (db < -45) {
      float v = (db - -60.0f) / 15.0f;
      *r = 0.0f;
      *g = 0.25f + 0.25f * v;
      *b = 0.5f - 0.5f * v;
    } else if (db < -30) {
      float v = (db - -45.0f) / 15.0f;
      *r = 0.75f * v;
      *g = 0.5f + 0.25f * v;
      *b = 0.0f;
    } else if (db < -0) {
      float v = (db - -30.0f) / 30.0f;
      *r = 0.75f + 0.25f * v;
      *g = 0.75f - 0.5f * v;
      *b = 0.25f * v;
    } else {
      *r = *g = *b = 1.0f;
    };

    if (*r < 0) *r = 0; if (*r > 1) *r = 1;
    if (*g < 0) *g = 0; if (*g > 1) *g = 1;
    if (*b < 0) *b = 0; if (*b > 1) *b = 1;

};

void user_audio_waterfall_append (float *values) {
  MUTEX_LOCK(audio.mutex);
  if (audio.initialized) {
    char *image = audio.buffer + 3 * audio.width * audio.index;
    for (int i = 0; i < audio.width; i++) {
      float r, g, b, db = 20.0f * logf(values[i]) / logf(10.0f);
      //db = -75.0 + 75.0 * i / audio.width;
      db_to_rgb(&r, &g, &b, db);
      image[3 * i + 0] = floor(r * 255.999);
      image[3 * i + 1] = floor(g * 255.999);
      image[3 * i + 2] = floor(b * 255.999);
    };
    int next_index = audio.index + 1;
    if (next_index >= audio.height) next_index = 0;
    audio.index = next_index;
  };
  MUTEX_UNLOCK(audio.mutex);
};

void user_signal_waterfall_append (float *values) {
  MUTEX_LOCK(signal.mutex);
  if (signal.initialized) {
    char *image = signal.buffer + 3 * signal.width * signal.index;
    for (int i = 0; i < signal.width; i++) {
      float r, g, b, db = 20.0f * logf(values[i]) / logf(10.0f);

      db += 10;

      //db = -75.0 + 75.0 * i / signal.width;
      db_to_rgb(&r, &g, &b, db);
      image[3 * i + 0] = floor(r * 255.999);
      image[3 * i + 1] = floor(g * 255.999);
      image[3 * i + 2] = floor(b * 255.999);
    };
    int next_index = signal.index + 1;
    if (next_index >= signal.height) next_index = 0;
    signal.index = next_index;
  };
  MUTEX_UNLOCK(signal.mutex);
};

static float last_spectrum_sum = 0.0f;

void user_waterfall_append (float *values) {
  if (waterfall_pause) return;
  MUTEX_LOCK(spectrum.mutex);
  if (spectrum.initialized) {
    char *image = spectrum.buffer + 3 * spectrum.width * spectrum.index;
    last_spectrum_sum = 0.0f;
    for (int i = 0; i < spectrum.width; i++) {
      last_spectrum_sum += values[i];
      float r, g, b, db = 20.0f * logf(values[i]) / logf(10.0f);
      db_to_rgb(&r, &g, &b, db);
      image[3 * i + 0] = floor(r * 255.999);
      image[3 * i + 1] = floor(g * 255.999);
      image[3 * i + 2] = floor(b * 255.999);
    };
    last_spectrum_sum /= spectrum.width;
    int next_index = spectrum.index + 1;
    if (next_index >= spectrum.height) next_index = 0;
    spectrum.index = next_index;
  };
  MUTEX_UNLOCK(spectrum.mutex);
};

/*
  char *waterfall_image = waterfall_buffer + 3 * 1024 * waterfall_index;

  for (int i = 0; i < 1024; i++) {
    float db = 20.0f * (logf(values[i]) / logf(10.0f));

    float r, g, b;
    db_to_rgb(&r, &g, &b, db);

    waterfall_image[3 * i + 0] = nearbyint(r * 255.99);
    waterfall_image[3 * i + 1] = nearbyint(g * 255.99);
    waterfall_image[3 * i + 2] = nearbyint(b * 255.99);
  };

  // Sometimes I wonder if I ought to use a mutex...
  int next_index = waterfall_index + 1;
  if (next_index >= 1024) next_index = 0;
  waterfall_index = next_index;

};*/

static int new_frequency = 0;
static int new_frequency_decimal_flag = 0;
static int new_frequency_decimate = 0;

static void raw_iq_waveform (void *parameter, struct window_data *window, struct event_data *event) {
  if (event && event->type == EVENT_KEY && event->action == GLFW_PRESS && event->key == GLFW_KEY_ESCAPE) window->close = 1;
  if (!event) {
    glColor3f(1.0, 1.0, 1.0);
    if (!option_my_gpu_sucks) glLineWidth(0.5);
    glBegin(GL_LINE_STRIP);
    for (int i = 0; i < 1024; i++) {
      glVertex2f(i + 0.5, 256.5 + iq_sample[2 * i + 0]);
    };
    glEnd();
    glBegin(GL_LINE_STRIP);
    for (int i = 0; i < 1024; i++) {
      glVertex2f(i + 0.5, 0.5 + iq_sample[2 * i + 1]);
    };
    glEnd();
  };
};

static void whatever (void *parameter, struct window_data *window, struct event_data *event) {
  if (event && event->type == EVENT_KEY && event->action == GLFW_PRESS && event->key == GLFW_KEY_ESCAPE) window->close = 1;

  #define HS 4

  static float reference[768] = {};
  static int use_reference = 0;

  static float data[HS][768] = {};
  static int last_index = 0;

  if (event && event->type == EVENT_KEY && event->action == GLFW_PRESS && event->key == GLFW_KEY_ENTER) {
    for (int i = 0; i < 768; i++) {
      reference[i] = data[HS-1][i];
    };
    use_reference = 1;
  };

  if (!event) {
    if (last_index != spectrum.index) {
      static int pretend_index = 0;
      float db = 20.0f * logf(last_spectrum_sum) / logf(10.0f);
      for (int i = 0; i < HS - 1; i++) {
        data[i][pretend_index] = data[i+1][pretend_index];
      };
      data[HS-1][pretend_index] = db;
      last_index = spectrum.index;
      pretend_index = last_index;
      radio_state.center_frequency = (25 + 2 * pretend_index) * 1000000;
      fprintf(stderr, "Scanning %0.0f MHz\n", radio_state.center_frequency / 1000000.0);
    };

    glBegin(GL_LINES);
    glColor3f(0.25, 0.25, 0.25);
    for (float y = 0; y < 256; y += 10.0 * 256 / 80.0) {
      glVertex2f(0, y);
      glVertex2f(768, y);
    };
    for (int f = 20; f <= 1540; f += 20) {
      float x = (f - 25) / 2 + 0.5f;
      if (f % 100 == 0) {
        glColor3f(0.5, 0.5, 0.5);
      } else {
        glColor3f(0.25, 0.25, 0.25);
      };
      glVertex2f(x, 0);
      glVertex2f(x, 256);
    };
    glEnd();

    for (int j = 0; j < HS; j++) {
      //float c = (j + 1) / (double) HS;
      //glColor3f(c, c, c);
      if (j == 0) glColor3f(0,0,1);
      if (j == 1) glColor3f(0,0.5,0);
      if (j == 2) glColor3f(1,0,0);
      if (j == 3) glColor3f(1,1,1);
      if (!option_my_gpu_sucks) glLineWidth(0.5);
      glBegin(GL_LINE_STRIP);
      for (int i = 0; i < 768; i++) {
        extern float alsa_audio_sample[];
        float y = 256.0f * (data[j][i] - -80.0f) / 80.0f;
        if (use_reference) {
          y = 128.0f + 256.0f * (data[j][i] - reference[i]) / 80.0f;
        };
        glVertex2f(i + 0.5, y);
      };
      glEnd();
    };


  };

};

#ifdef TEST

#include "filter.h"
#include "better_fir_filters.c"

struct moofilter {
  float frequency;
  float response;
  float magnitude;
  float phase_offset;
  complex double phase;
  complex double state;
  int enabled;
};

struct moofilter moof[17] = {};
struct moofilter copy[17] = {};

static void bullshit (void *parameter, struct window_data *window, struct event_data *event) {
  if (event && event->type == EVENT_KEY && event->action == GLFW_PRESS && event->key == GLFW_KEY_ESCAPE) window->close = 1;

  static double best_score = 1e+50;
  static float desired[1000];

  const float floor = -75;

  static int initialized = 0;
  if (!initialized) {
    initialized = 1;
    for (int i = 0; i < 17; i++) {
      moof[i].frequency = (i - 8) / 48.0 + 0.5f;
      moof[i].magnitude = 0.0f;
      moof[i].response = -10.0f;
      moof[i].phase = 1.0f;
      moof[i].state = 0.0f;
      if (i >= 5 && i <= 11) {
        moof[i].enabled = 1;
        moof[i].magnitude = 0.5f;
      };
      if (i == 8) moof[i].magnitude = 1.0;
    };
    for (int i = 0; i < 1000; i++) {
      float db = floor;
      float f = 24.0f * i / 1000.0f;
      #if 1
      if (f >= 6 && f < 9) {
        db = 1/0.0;
      } else if (f >= 9 && f <= 15) {
        db = 0;
      } else if (f > 15 && f <= 18) {
        db = 1/0.0;
      };
      #else
      if (f >= 9 && f <= 15) db = 0;
      #endif
      desired[i] = powf(10.0f, db / 20.0f);
    };
  };

  int keep = 0;
  int index = 0;
  for (int i = 0; i < 17; i++) {
    int y = 516 + 16 * i;

    float f = 24 * moof[i].frequency;
    keep += gui_slider(event, index++, 0, y, 250, 16, 0.0, 24.0, &f);
    //f = nearbyint(f * 4.0) / 4.0;
    gui_printf(event, 64, y, "%f", f);
    moof[i].frequency = f / 24.0;

    float m = moof[i].magnitude;
    keep += gui_slider(event, index++, 256, y, 250, 16, -3.0, +3.0, &m);
    //m = nearbyint(m * 10) / 10;
    gui_printf(event, 256 + 64, y, "%f", m);
    moof[i].magnitude = m;

    float r = moof[i].response;
    keep += gui_slider(event, index++, 512, y, 250, 16, -40.0, 0.0, &r);
    //r = nearbyint(r);
    gui_printf(event, 512 + 64, y, "%f", r);
    moof[i].response = r;

    float p = moof[i].phase_offset;
    keep += gui_slider(event, index++, 768, y, 125, 16, -0.5, +0.5, &p);
    //p = nearbyint(p * 10) / 10;
    gui_printf(event, 768 + 64, y, "%f", p);
    moof[i].phase_offset = p;

    if (moof[i].enabled) {
      if (gui_link(event, index++, 768 + 140, y, "Disable")) moof[i].enabled = 0, keep++;
    } else {
      if (gui_link(event, index++, 768 + 140, y, "Enable")) moof[i].enabled = 1, keep++;
    };

  };

  static int sweep_length = 400;
  static int step_size = 5;

  int y = 516 + 16 * 18;
  static float sl = 400, ss = 5;
  if (gui_slider(event, index++, 0, y, 250, 16, 100, 4000, &sl)) {
    sweep_length = nearbyint(sl);
    printf("sweep length: %d\n", sweep_length);
  };
  if (gui_slider(event, index++, 256, y, 250, 16, 1, 10, &ss)) {
    step_size = nearbyint(ss);
    printf("step size: %d\n", step_size);
  };


  if (keep) best_score = 1e50;

  // 1000 pixels wide, 500 pixels tall
  // 25 pixels per 10 db vertical
  // 500 pixels = +40 db
  // 400 pixels = 0 db
  // 300 pixels = -40
  // 200 = -80
  // 100 = -120
  // 0 = -160

  #define line(s, t, u, v) glVertex2f(s, t); glVertex2f(u, v)
  #define db2y(db) (400 + db * 2.5f)

  const int divisions = 24;

  if (!event) {
    color(9, -5);
    glBegin(GL_LINES);
    for (int f = 0; f <= divisions; f++) {
      float x = 1000 * f / (double) divisions;
      line(x, 0, x, 500);
    };
    glEnd();
    for (int f = 0; f <= divisions; f++) {
      float x = 1000 * f / (double) divisions;
      if (f >= 10) x -= 4;
      gui_printf(event, x - 4, 500, "%d", f);
    };
    glBegin(GL_LINES);
    for (int db = 40; db >= -180; db -= 10) {
      if (db == 0 || db == -40 || db == -80) {
        color(2, -5);
      } else {
        color(9, -5);
      };
      float y = db2y(db);
      line(0, y, 1000, y);
    };
    glEnd();
    for (int db = 40; db >= -180; db -= 10) {
      if (db == 0 || db == -40 || db == -80) {
        color(2, -5);
      } else {
        color(9, -5);
      };
      float y = db2y(db);
      gui_printf(event, 1004, y - 8, "%+03d dB", db);
    };

    static struct filter_lowpass filter[4] = {};
    static int sweep = 0;
    static float results[1000] = {};
    static float best_results[1000];

    for (int j = 0; j < 100; j++) {

      complex double signal;
      for (int i = 0; i < sweep_length; i++) {
        signal = cexp(2 * M_PI * I * (sweep / 1000.0) * i);
        //for (int j = 0; j < 1; j++) {
        //  //signal = filter_lowpass(&filter[j], 9, 0.5 / 10, signal);
        //  signal = fir_filter_2400_200 (signal, 0);
        //};

        complex double output = 0.0;
        for (int f = 0; f < 16; f++) {
          if (!moof[f].enabled) continue;
          moof[f].phase *= cexp(2 * M_PI * I * moof[f].frequency);
          moof[f].phase /= cabs(moof[f].phase);
          complex double t = signal / moof[f].phase;
          float m = moof[f].magnitude;
          float r = pow(10.0, moof[f].response / 20.0);
          moof[f].state *= (1.0f - r);
          moof[f].state += r * t;
          output += moof[f].phase * m * moof[f].state * cexp(2 * M_PI * I * moof[f].phase_offset);
        };
        signal = output;

      };
      results[sweep] = cabsf(signal);
      sweep += step_size; if (sweep >= 1000) sweep = 0;

      if (sweep == 0) {
        double ms = 0.0f;
        int count = 0;
        for (int i = 0; i < 1000; i += step_size) {
          if (desired[i] != 1/0.0) {
            float c = 20.0f * logf(results[i]) / logf(10.0f);
            if (c < floor) c = floor;
            float d = c - 20.0f * logf(desired[i]) / logf(10.0f);
            ms += d * d;
            count++;
          };
        };
        ms /= count;
        double score = sqrt(ms);

        //static float evolution_rate = 1.0f;
        static float er[16][4];
        if (er[0][0] == 0) {
          for (int x = 0; x < 16; x++) {
            for (int y = 0; y < 4; y++) {
              er[x][y] = 1.0;
            };
          };
        };

        #define evolution_rate er[which][what]

        static int tries[16][4] = {};

        static int which = 0;
        static int what = 8;
        if (best_score > score) {
          printf("new best score: %0.18f is better than %0.18f\n", score, best_score);
          best_score = score;
          memmove(copy, moof, sizeof(moof));
          memmove(best_results, results, sizeof(results));
          if (tries[which][what] < 5) {
            evolution_rate *= 1.1;
            if (evolution_rate > 10) evolution_rate = 10;
            printf("evolution rate is now %f\n", evolution_rate);
          };
          tries[which][what] = 0;
        } else {
          memmove(moof, copy, sizeof(moof));
          tries[which][what]++;
          if (tries[which][what] > 10) {
            evolution_rate *= 0.5;
            if (evolution_rate < 0.0001) evolution_rate = 0.0001;
            tries[which][what] = 0;
            printf("evolution rate is now %f\n", evolution_rate);
          } else if (tries[which][what] > 5) {
            what = rand() & 3;
            do {
              which = 8 - (rand() & 7);
            } while (!moof[which].enabled);
          };
        };
        for (int f = 0; f <= 8; f++) {
          if (f != which) continue;
          if (!moof[f].enabled) continue;
          #define randy ((rand() / (double) RAND_MAX) - 0.5)
          if (what == 0 && f != 8) moof[f].frequency += randy * 0.1 * evolution_rate;
          if (what == 1) moof[f].magnitude += randy * 0.01 * evolution_rate;
          if (what == 2) moof[f].response += randy * 1.0 * evolution_rate;
          if (moof[f].response > 0) moof[f].response = 0;
          if (moof[f].response < -40) moof[f].response = -40;

          if (what == 3) moof[f].phase_offset += randy * 0.1 * evolution_rate;
          if (moof[f].phase_offset > 0.5) moof[f].phase_offset -= 1.0;
          if (moof[f].phase_offset < -0.5) moof[f].phase_offset += 1.0;

        };

        // some symmetry seems obvious
        moof[8].phase_offset = 0;
        for (int i = 1; i <= 7; i++) {
          moof[8 + i].frequency = 1.0 - moof[8 - i].frequency;
          moof[8 + i].magnitude = moof[8 - i].magnitude;
          moof[8 + i].response = moof[8 - i].response;
          moof[8 + i].phase_offset = -moof[8 - i].phase_offset;
          moof[8 + i].enabled = -moof[8 - i].enabled;
        };
      };


    };

    color(6, 0);
    glBegin(GL_LINE_STRIP);
    for (int x = 0; x < 1000; x += step_size) {
      float db = 20.0f * logf(desired[x]) / logf(10.0f);
      //if (x < 100) printf("%d = %f, sweep=%d\n", x, results[x], sweep);
      float y = db2y(db);
      glVertex2f(x, y);
    };
    glEnd();

    color(3, 0);
    glBegin(GL_LINE_STRIP);
    for (int x = 0; x < 1000; x += step_size) {
      float db = 20.0f * logf(best_results[x]) / logf(10.0f);
      //if (x < 100) printf("%d = %f, sweep=%d\n", x, results[x], sweep);
      float y = db2y(db);
      glVertex2f(x, y);
    };
    glEnd();

    color(5, 0);
    glBegin(GL_LINE_STRIP);
    for (int x = 0; x < 1000; x += step_size) {
      float db = 20.0f * logf(results[x]) / logf(10.0f);
      //if (x < 100) printf("%d = %f, sweep=%d\n", x, results[x], sweep);
      float y = db2y(db);
      glVertex2f(x, y);
    };
    glEnd();

  };

};

#endif

static void raw_audio_waveform (void *parameter, struct window_data *window, struct event_data *event) {
  if (event && event->type == EVENT_KEY && event->action == GLFW_PRESS && event->key == GLFW_KEY_ESCAPE) window->close = 1;
  if (!event) {
    glColor3f(1.0, 1.0, 1.0);
    if (!option_my_gpu_sucks) glLineWidth(0.5);
    glBegin(GL_LINE_STRIP);
    for (int i = 0; i < 1024; i++) {
      extern float alsa_audio_sample[];
      glVertex2f(i + 0.5, 128 + 128 * alsa_audio_sample[i]);
    };
    glEnd();
  };
};

static void smooth(int state) {
  if (option_my_gpu_sucks) state = 0;
  if (state) {
    glEnable(GL_BLEND);
    glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
    glHint(GL_POINT_SMOOTH_HINT, GL_NICEST);
    glHint(GL_LINE_SMOOTH_HINT, GL_NICEST);
    glHint(GL_POLYGON_SMOOTH_HINT, GL_NICEST);
    glEnable(GL_POINT_SMOOTH);
    glEnable(GL_LINE_SMOOTH);
    glEnable(GL_POLYGON_SMOOTH);
  } else {
    glDisable(GL_BLEND);
    glDisable(GL_POINT_SMOOTH);
    glDisable(GL_LINE_SMOOTH);
    glDisable(GL_POLYGON_SMOOTH);
  };
};

static void ui (void *parameter, struct window_data *window, struct event_data *event) {

  if (option_my_gpu_sucks) {
    if (!event) {
      static double last = 0;
      double now = easy_time();
      double delay = 0.1 - (now - last);
      //fprintf(stderr, "Interval: %f, delay %f\n", now - last, delay);
      easy_sleep(delay);
      last = easy_time();
    };
  };

  int waterfall_width = 1024;
  int waterfall_height = 768;
  int waterfall_x = window->width - waterfall_width;
  int waterfall_y = 0;

  static float noise_level = 0.00025f;

  if (event) {
    if (event->type == EVENT_KEY && event->action == GLFW_PRESS) {
      if (event->key == GLFW_KEY_ESCAPE) window->close = 1;
      if (event->key == GLFW_KEY_ENTER || event->key == GLFW_KEY_KP_ENTER) {
        if (new_frequency) {
          radio_state.center_frequency = (new_frequency / (double) new_frequency_decimate + nearbyint(radio_state.upconverter_frequency / 1000000.0)) * 1000000.0;
          new_frequency = 0;
          new_frequency_decimal_flag = 0;
          new_frequency_decimate = 1;
        };
      };
    };
    int tuning_step = 1;
    if (radio_state.mode == RADIO_MODE_LSB || radio_state.mode == RADIO_MODE_USB) {
      tuning_step = 5000;
    } else if (radio_state.mode == RADIO_MODE_AM_10) {
      tuning_step = 10000;
    } else if (radio_state.mode == RADIO_MODE_FM_12) {
      tuning_step = 12500;
    } else if (radio_state.mode == RADIO_MODE_FM_25) {
      tuning_step = 25000;
    } else if (radio_state.mode == RADIO_MODE_FM_200) {
      tuning_step = 200000;
    };
    if (event->modifiers & GLFW_MOD_SHIFT) {
      tuning_step /= 10;
    };
    if (event->modifiers & GLFW_MOD_CONTROL) {
      tuning_step /= 100;
    };
    if (event->type == EVENT_KEY && (event->action == GLFW_PRESS || event->action == GLFW_REPEAT)) {
      //if (event->modifiers & GLFW_MOD_SHIFT) {
      //  if (event->key == GLFW_KEY_LEFT) {
      //    radio_state.center_frequency -= 100000;
      //    radio_state.offset_frequency += 100000;
      //  };
      //  if (event->key == GLFW_KEY_RIGHT) {
      //    radio_state.center_frequency += 100000;
      //    radio_state.offset_frequency -= 100000;
      //  };
      //} else {
      if (event->key == GLFW_KEY_LEFT) {
        radio_state.offset_frequency -= tuning_step;
      };
      if (event->key == GLFW_KEY_RIGHT) {
        radio_state.offset_frequency += tuning_step;
      };
      //};
      if (event->key == GLFW_KEY_UP) {
        float x, y;
        radio_get_playback_position(&x, &y);
        y -= 4; if (y < 0) y = 0;
        radio_set_time_offset(y);
      };
      if (event->key == GLFW_KEY_DOWN) {
        float x, y;
        radio_get_playback_position(&x, &y);
        y += 4; if (y > 767) y = 767;
        radio_set_time_offset(y);
      };
      if (event->key == '[' && event->action != GLFW_REPEAT) {
        radio_state.center_frequency -= 2000000;
      };
      if (event->key == ']' && event->action != GLFW_REPEAT) {
        radio_state.center_frequency += 2000000;
      };
    };
    if (event->type == EVENT_CHARACTER) {
      //fprintf(stderr, "Character: %d, (%08x)\n", event->character, event->character);
      if (event->character >= '0' && event->character <= '9') {
        new_frequency *= 10;
        new_frequency += event->character - '0';
        if (new_frequency_decimal_flag) {
          new_frequency_decimate *= 10;
        } else {
          new_frequency_decimate = 1;
        };
        fprintf(stderr, "New Frequency: %0.6f MHz\n", new_frequency / (double) new_frequency_decimate);
      };
      if (event->character == '.') {
        new_frequency_decimal_flag = 1;
        new_frequency_decimate = 1;
      };
      if (event->character == 'f') radio_state.mode = RADIO_MODE_FM_12;
      if (event->character == 'a') radio_state.mode = RADIO_MODE_AM_10;
      if (event->character == 'l') radio_state.mode = RADIO_MODE_LSB;
      if (event->character == 'u') radio_state.mode = RADIO_MODE_USB;
      if (event->character == 's') radio_set_squelch();
      if (event->character == 'i') radio_start_iq_recording();
      if (event->character == 'z') noise_level = sqrt(radio_signal_strength_squared);
      static int recording = 0;
      if (event->character == 'r') {
        if (recording) {
          radio_stop_recording();
          recording = 0;
        } else {
          radio_start_recording();
          recording = 1;
        };
      };
    };
    if (event->type == EVENT_BUTTON && event->action == GLFW_PRESS && event->button == 1) {
      if (event->x >= waterfall_x && event->x < waterfall_x + waterfall_width && event->y >= waterfall_y && event->y < waterfall_y + waterfall_height) {
        radio_demodulator_enable = 0;
      };
    };
    static int waterfall_click = 0;
    if (event->type == EVENT_BUTTON && event->action == GLFW_PRESS && event->button == 0) {
      if (event->x >= waterfall_x && event->x < waterfall_x + waterfall_width && event->y >= waterfall_y && event->y < waterfall_y + waterfall_height) {
        waterfall_click = 1;
        radio_scan_enable = 0;
        radio_set_time_offset(767 - (event->y - waterfall_y));
        radio_demodulator_enable = 1;
      }
    };
    if (event->type == EVENT_BUTTON && event->action == GLFW_RELEASE && event->button == 0) {
      waterfall_click = 0;
    };
    if (waterfall_click) {
      int x = event->x - waterfall_x;
      if (x < 0) x = 0; if (x > waterfall_width - 1) x = waterfall_width - 1;
      int uc_offset = radio_state.upconverter_frequency - nearbyint(radio_state.upconverter_frequency / 500.0) * 500;
      radio_state.offset_frequency = nearbyint((x - 512) / (float) waterfall_width * IQ_SAMPLE_RATE / 500.0) * 500 + uc_offset;
    };
    if (event->type == EVENT_SCROLL) {
      radio_state.offset_frequency += 500 * event->dy;
    };
  } else {

    // Draw the main waterfall...

    if (1) {
      MUTEX_LOCK(spectrum.mutex);
      if (!spectrum.initialized) {
        memory_allocate(spectrum.texture, spectrum.height * sizeof(GLuint));
        spectrum.initialized = 1;
        glGenTextures(spectrum.height, spectrum.texture);
        for (int i = 0; i < spectrum.height; i++) {
          glBindTexture(GL_TEXTURE_2D, spectrum.texture[i]);
          glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
          glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
          glTexImage2D(GL_TEXTURE_2D, 0, GL_RGB, spectrum.width, 1, 0, GL_RGB, GL_UNSIGNED_BYTE, spectrum.buffer + 3 * spectrum.width * (spectrum.height - 1));
        };
      };

      static int last_index = 0;
      while (last_index != spectrum.index) {
        glBindTexture(GL_TEXTURE_2D, spectrum.texture[last_index]);
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
        glTexImage2D(GL_TEXTURE_2D, 0, GL_RGB, spectrum.width, 1, 0, GL_RGB, GL_UNSIGNED_BYTE, spectrum.buffer + 3 * spectrum.width * last_index);
        last_index++; if (last_index >= spectrum.height) last_index = 0;
      };

      int draw_height = 768;
      int wi = last_index;
      wi += spectrum.height - draw_height;
      if (wi >= spectrum.height) wi -= spectrum.height;

      glEnable(GL_TEXTURE_2D);
      glColor4f(1.0, 1.0, 1.0, 1.0);

      for (int y = 0; y < draw_height; y++) {
        glBindTexture(GL_TEXTURE_2D, spectrum.texture[wi]);
        glBegin(GL_QUADS);
        glTexCoord2d(0.0f, 0.0f); glVertex2f(256, y);
        glTexCoord2d(1.0f, 0.0f); glVertex2f(256 + spectrum.width, y);
        glTexCoord2d(1.0f, 1.0f); glVertex2f(256 + spectrum.width, y + 1);
        glTexCoord2d(0.0f, 1.0f); glVertex2f(256, y + 1);
        glEnd();
        if (++wi >= spectrum.height) wi = 0;
      };
      glDisable(GL_TEXTURE_2D);
      MUTEX_UNLOCK(spectrum.mutex);
    };

    /*{
      static int initialized = 0;
      if (!initialized) {
        initialized = 1;
        glGenTextures(1024, waterfall_texture);
        for (int i = 0; i < 1024; i++) {
          glBindTexture(GL_TEXTURE_2D, waterfall_texture[i]);
          glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
          glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
          glTexImage2D(GL_TEXTURE_2D, 0, GL_RGB, 1024, 1, 0, GL_RGB, GL_UNSIGNED_BYTE, waterfall_buffer + 3 * 1024 * 1023);
        };
      };

      static int last_index = 0;
      while (last_index != waterfall_index) {
        glBindTexture(GL_TEXTURE_2D, waterfall_texture[last_index]);
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
        glTexImage2D(GL_TEXTURE_2D, 0, GL_RGB, 1024, 1, 0, GL_RGB, GL_UNSIGNED_BYTE, waterfall_buffer + 3 * 1024 * last_index);
        last_index++; if (last_index >= 1024) last_index = 0;
      };

      int wi = waterfall_index;
      wi += 1024 - waterfall_height;
      if (wi > 1024) wi -= 1024;

      glEnable(GL_TEXTURE_2D);
      glColor4f(1.0, 1.0, 1.0, 1.0);
      for (int y = 0; y < waterfall_height; y++) {
        glBindTexture(GL_TEXTURE_2D, waterfall_texture[wi]);
        glBegin(GL_QUADS);
        glTexCoord2d(0.0f, 1.0f); glVertex2f(waterfall_x, y);
        glTexCoord2d(1.0f, 0.0f); glVertex2f(waterfall_x + 1024, y);
        glTexCoord2d(1.0f, 1.0f); glVertex2f(waterfall_x + 1024, y + 1);
        glTexCoord2d(0.0f, 1.0f); glVertex2f(waterfall_x, y + 1);
        glEnd();
        if (++wi >= 1024) wi = 0;
      };
      glDisable(GL_TEXTURE_2D);
    };*/

    if (!option_less_stuff) {

      // Draw the audio waterfall...

      if (1) {
        MUTEX_LOCK(audio.mutex);
        if (!audio.initialized) {
          memory_allocate(audio.texture, audio.height * sizeof(GLuint));
          audio.initialized = 1;
          glGenTextures(audio.height, audio.texture);
          for (int i = 0; i < audio.height; i++) {
            glBindTexture(GL_TEXTURE_2D, audio.texture[i]);
            glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
            glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
            glTexImage2D(GL_TEXTURE_2D, 0, GL_RGB, audio.width, 1, 0, GL_RGB, GL_UNSIGNED_BYTE, audio.buffer + 3 * audio.width * (audio.height - 1));
          };
        };

        static int last_index = 0;
        while (last_index != audio.index) {
          glBindTexture(GL_TEXTURE_2D, audio.texture[last_index]);
          glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
          glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
          glTexImage2D(GL_TEXTURE_2D, 0, GL_RGB, audio.width, 1, 0, GL_RGB, GL_UNSIGNED_BYTE, audio.buffer + 3 * audio.width * last_index);
          last_index++; if (last_index >= audio.height) last_index = 0;
        };

        int draw_height = 192;
        int wi = last_index;
        wi += audio.height - draw_height;
        if (wi >= audio.height) wi -= audio.height;

        glEnable(GL_TEXTURE_2D);
        glColor4f(1.0, 1.0, 1.0, 1.0);

        for (int x = 0; x < draw_height; x++) {
          glBindTexture(GL_TEXTURE_2D, audio.texture[wi]);
          glBegin(GL_QUADS);
          glTexCoord2d(0.0f, 0.0f); glVertex2f(x, 0);
          glTexCoord2d(1.0f, 0.0f); glVertex2f(x, 0 + audio.width);
          glTexCoord2d(1.0f, 1.0f); glVertex2f(x + 1, 0 + audio.width);
          glTexCoord2d(0.0f, 1.0f); glVertex2f(x + 1, 0);
          glEnd();
          if (++wi >= audio.height) wi = 0;
        };

        glDisable(GL_TEXTURE_2D);

        smooth(1);
        glColor3f(1.0, 1.0, 1.0);
        glBegin(GL_LINES);
        float x = 192;
        for (int i = 0; i <= 24; i++) {
          float w;
          if (i % 5 == 0) {
            if (!option_my_gpu_sucks) glLineWidth(1.5);
            w = 8;
          } else {
            if (!option_my_gpu_sucks) glLineWidth(0.5);
            w = 4;
          };
          float y = i * 1000.0 * 256 / 24000.0;
          glVertex2f(x, y);
          glVertex2f(x + w, y);
        };
        glEnd();
        smooth(0);
        for (int i = 0; i <= 24; i += 5) {
          float y = i * 1000.0 * 256 / 24000.0;
          if (i == 0) y += 8;
          gui_printf(event, x + 12, nearbyint(y - 8), "%d kHz", i);
        };
        MUTEX_UNLOCK(audio.mutex);
      };

      // Draw the signal waterfall...

      if (radio_state.mode != RADIO_MODE_FM_200) {
        MUTEX_LOCK(signal.mutex);
        if (!signal.initialized) {
          memory_allocate(signal.texture, signal.height * sizeof(GLuint));
          signal.initialized = 1;
          glGenTextures(signal.height, signal.texture);
          for (int i = 0; i < signal.height; i++) {
            glBindTexture(GL_TEXTURE_2D, signal.texture[i]);
            glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
            glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
            glTexImage2D(GL_TEXTURE_2D, 0, GL_RGB, signal.width, 1, 0, GL_RGB, GL_UNSIGNED_BYTE, signal.buffer + 3 * signal.width * (signal.height - 1));
          };
        };

        static int last_index = 0;
        while (last_index != signal.index) {
          glBindTexture(GL_TEXTURE_2D, signal.texture[last_index]);
          glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
          glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
          glTexImage2D(GL_TEXTURE_2D, 0, GL_RGB, signal.width, 1, 0, GL_RGB, GL_UNSIGNED_BYTE, signal.buffer + 3 * signal.width * last_index);
          last_index++; if (last_index >= signal.height) last_index = 0;
        };

        int draw_height = 128;
        int wi = last_index;
        wi += signal.height - draw_height;
        if (wi >= signal.height) wi -= signal.height;

        glEnable(GL_TEXTURE_2D);
        glColor4f(1.0, 1.0, 1.0, 1.0);

        for (int y = 0; y < draw_height; y++) {
          glBindTexture(GL_TEXTURE_2D, signal.texture[wi]);
          glBegin(GL_QUADS);
          glTexCoord2d(0.0f, 0.0f); glVertex2f(0, y + 256);
          glTexCoord2d(1.0f, 0.0f); glVertex2f(0 + signal.width, y + 256);
          glTexCoord2d(1.0f, 1.0f); glVertex2f(0 + signal.width, y + 256 + 1);
          glTexCoord2d(0.0f, 1.0f); glVertex2f(0, y + 256 + 1);
          glEnd();
          if (++wi >= signal.height) wi = 0;
        };
        glDisable(GL_TEXTURE_2D);
        MUTEX_UNLOCK(signal.mutex);
      };

    };

    if (radio_demodulator_enable) {

      float px, py;
      radio_get_playback_position(&px, &py);
      px *= 1024.0f; py = 768 - py;

      smooth(1);

      glBegin(GL_TRIANGLES);
      glColor4f(1.0, 1.0, 1.0, 1.0);
      glVertex2f(waterfall_x + px, waterfall_y + py);
      glVertex2f(waterfall_x + px - 5, waterfall_y + py - 20);
      glVertex2f(waterfall_x + px + 5, waterfall_y + py - 20);
      glEnd();

      glBegin(GL_LINE_LOOP);
      glColor4f(0.0, 0.0, 0.0, 1.0);
      glVertex2f(waterfall_x + px, waterfall_y + py);
      glVertex2f(waterfall_x + px - 5, waterfall_y + py - 20);
      glVertex2f(waterfall_x + px + 5, waterfall_y + py - 20);
      glEnd();

      smooth(0);

    };

  };

  int index = 0;
  int x = 4;
  int y = window->height - 4;
  glColor3f(1.0, 1.0, 1.0);
  gui_printf(event, x, y -= 16, "Center Frequency: %0.6f MHz", (radio_state.center_frequency - radio_state.upconverter_frequency) / 1000000.0);
  gui_printf(event, x, y -= 16, "Tuned Frequency: %0.6f MHz", (radio_state.center_frequency + radio_state.offset_frequency - radio_state.upconverter_frequency) / 1000000.0);
  y -= 4;

  y -= 16;
  gui_printf(event, x, y, "Direct Sampling: ");
  gui_radio(event, index++, x + 18 * 8, y, "Off", &radio_state.direct_sampling, 0);
  gui_radio(event, index++, x + 23 * 8, y, "I", &radio_state.direct_sampling, 1);
  gui_radio(event, index++, x + 26 * 8, y, "Q", &radio_state.direct_sampling, 2);
  y -= 4;

  y -= 16;
  gui_printf(event, x, y, "Up/Down Converter: ");
  gui_radio(event, index++, x + 20 * 8, y, "On", &radio_state.upconverter_frequency, option_upconverter_frequency);
  gui_radio(event, index++, x + 24 * 8, y, "Off", &radio_state.upconverter_frequency, 0);
  y -= 4;

  y -= 16;
  glColor3f(1.0, 1.0, 1.0);
  gui_printf(event, x, y, "Gain:");
  float gain = radio_state.gain_index + 0.5f;
  if (radio_state.gain_mode == 0) gain = 0.0f;
  if (gui_slider(event, index++, x + 6 * 8, y, 256 - 14 * 8, 16, 0.0, 29.0, &gain)) {
    int g = floor(gain);
    if (g < 0) g = 0;
    if (g > 28) g = 28;
    radio_state.gain_mode = 1;
    radio_state.gain_index = g;
  };
  if (radio_state.gain_mode == 0) {
    color(5, 0);
    gui_printf(event, x + 84, y, "Automatic");
  } else {
    float values[] = {0.0, 0.9, 1.4, 2.7, 3.7, 7.7, 8.7, 12.5, 14.4, 15.7, 16.6, 19.7, 20.7, 22.9, 25.4, 28.0, 29.7, 32.8, 33.8, 36.4, 37.2, 38.6, 40.2, 42.1, 43.4, 43.9, 44.5, 48.0, 49.6};
    color(9, 9);
    gui_printf(event, x + 92, y, "%4.1f dB", values[radio_state.gain_index]);
  };
  if (gui_radio(event, index++, x + 256 - 6 * 8, y, "Auto", &radio_state.gain_mode, 0)) {
    radio_state.gain_mode = 0;
    gain = 0.0f;
  };
  y -= 4;

  int tx;

  gui_printf(event, x, y -= 16, "Demodulation Mode:");
  tx = x + 16; y -= 8;                   ////////////////////////////////
  gui_radio(event, index++, tx, y -= 20, "USB (5 kHz channels)", &radio_state.mode, RADIO_MODE_USB);
  gui_radio(event, index++, tx, y -= 20, "LSB (5 kHz channels)", &radio_state.mode, RADIO_MODE_LSB);
  gui_radio(event, index++, tx, y -= 20, "AM (10 kHz channels)", &radio_state.mode, RADIO_MODE_AM_10);
  gui_radio(event, index++, tx, y -= 20, "FM (12.5 kHz channels)", &radio_state.mode, RADIO_MODE_FM_12);
  gui_radio(event, index++, tx, y -= 20, "FM (25 kHz channels)", &radio_state.mode, RADIO_MODE_FM_25);
  gui_radio(event, index++, tx, y -= 20, "FM (200 kHz broadcast)", &radio_state.mode, RADIO_MODE_FM_200);
  y -= 12;

  y -= 16;
  gui_printf(event, x, y, "Voice Inversion: ");
  gui_radio(event, index++, x + 20 * 8, y, "On", &radio_invert_audio, 1);
  gui_radio(event, index++, x + 24 * 8, y, "Off", &radio_invert_audio, 0);
  y -= 4;

  y -= 16;
  {
    glColor3f(1.0, 1.0, 1.0);
    gui_printf(event, x, y, "Static:");
    float gain = -radio_audio_gain;
    if (gui_slider(event, index++, x + 8 * 8, y, 256 - 9 * 8, 16, 0.0, 72.0, &gain)) {
      radio_audio_gain = -gain;
    };
  };
  y -= 4;

  x += 8;

  if (gui_link(event, index++, x, y -= 16, "Enable Scan Mode (kinda lame)")) {
    radio_scan_enable = 1;
    radio_demodulator_enable = 1;
    radio_set_time_offset(0);
  };
  y -= 8;

  if (gui_link(event, index++, x, y -= 16, "View Raw IQ Waveform")) {
    static volatile int thread_run_flag = 0;
    if (thread_run_flag == 0) {
      thread_run_flag = 1;
      glfw_generic_event_loop_thread(1024, 512, "Raw IQ Waveform", 0, &thread_run_flag, raw_iq_waveform, NULL);
    };
  };
  y -= 8;

  if (gui_link(event, index++, x, y -= 16, "View Audio Waveform")) {
    static volatile int thread_run_flag = 0;
    if (thread_run_flag == 0) {
      thread_run_flag = 1;
      glfw_generic_event_loop_thread(1024, 256, "Audio Waveform", 0, &thread_run_flag, raw_audio_waveform, NULL);
    };
  };
  y -= 8;

  /* No one wants this stupid thing...
  if (gui_link(event, index++, x, y -= 16, "Stupid Frequency Scanner")) {
    static volatile int thread_run_flag = 0;
    if (thread_run_flag == 0) {
      thread_run_flag = 1;
      glfw_generic_event_loop_thread(768, 256, "Whatever", 0, &thread_run_flag, whatever, NULL);
    };
  };
  y -= 8;
  */

  #ifdef TEST

  if (gui_link(event, index++, x, y -= 16, "Filter Test Window")) {
    static volatile int thread_run_flag = 0;
    if (thread_run_flag == 0) {
      thread_run_flag = 1;
      glfw_generic_event_loop_thread(1064, 564 + 256, "Bullshit", 0, &thread_run_flag, bullshit, NULL);
    };
  };
  y -= 8;

  if (gui_link(event, index++, x, y -= 16, "Pause Waterfall")) waterfall_pause = !waterfall_pause;
  y -= 8;

  #endif

  gui_printf(event, x + 18 * 8, y - 16, "flag = %d", radio_test_flag);
  if (gui_link(event, index++, x, y -= 16, "Toggle Test Flag")) radio_test_flag = !radio_test_flag;
  y -= 8;

  // Draw the IQ signal strength meter...

  x = 256; y = window->height - 72;
  if (!event) {
    smooth(1);
    if (!option_my_gpu_sucks) glLineWidth(1.5);
    glBegin(GL_LINES);
    for (float db = 0; db >= -48; db -= (option_my_gpu_sucks ? 2.0 : 0.2)) {
      float r, g, b;
      db_to_rgb(&r, &g, &b, db);
      glColor3f(r, g, b);
      float angle = 30.0f - 120.0f * db / 48.0f;
      if (angle > 150.0f) angle = 150.0f;
      float w = M_PI * angle / 180.0f;
      glVertex2f(x + 64 + 60 * cos(w), y + 0 + 60 * sin(w));
      glVertex2f(x + 64 + 64 * cos(w), y + 0 + 64 * sin(w));
    };
    glColor3f(1.0, 1.0, 1.0);
    for (int db = 0; db >= -48; db -= 10) {
      float angle = 30.0f - 120.0f * db / 48.0f;
      if (angle > 150.0f) angle = 150.0f;
      float w = M_PI * angle / 180.0f;
      glVertex2f(x + 64 + 60 * cos(w), y + 0 + 60 * sin(w));
      glVertex2f(x + 64 + 64 * cos(w), y + 0 + 64 * sin(w));
    };
    glColor3f(1.0, 0.2, 0.2);
    glVertex2f(x + 64, y + 0);
    float db = 20.0f * logf(sqrtf(radio_iq_strength_squared)) / logf(10.0f);
    // full scale (0 db) = 60 degrees right
    float angle = 30.0f - 120.0f * db / 48.0f;
    if (angle > 150.0f) angle = 150.0f;
    float w = M_PI * angle / 180.0f;
    glVertex2f(x + 64 + 58 * cos(w), y + 0 + 58 * sin(w));
    glEnd();
    smooth(0);
    static float average = -80.0f;
    average *= 0.9; average += 0.1 * db;
    if (average < -80.0f) average = -80.0f;
    glColor3f(1.0, 1.0, 1.0);
    gui_printf(event, x + 8, y + 0, "IQ Input Level");
    gui_printf(event, x + 32, y + 16, "%3.1f dB", average);
  };

  // Draw the SNR meter...

  x = window->width - 128; y = window->height - 72;
  if (!event) {
    smooth(1);
    if (!option_my_gpu_sucks) glLineWidth(1.5);
    glBegin(GL_LINES);
    glColor3f(1.0, 1.0, 1.0);
    for (int db = -10; db <= 90; db += 5) {
      float angle = 150.0f - 120.0f * (db + 10) / 90.0f;
      if (angle > 150.0f) angle = 150.0f;
      if (angle < 30.0f) angle = 30.0f;
      float w = M_PI * angle / 180.0f;
      float l = 4 + 4 * (db % 10 == 0);
      if (db == 0) {
        glColor3f(1.0, 0.2, 0.2);
      } else if (db == 5) {
        glColor3f(1.0, 1.0, 1.0);
      };
      glVertex2f(x + 64 + (64 - l) * cos(w), y + 0 + (64 - l) * sin(w));
      glVertex2f(x + 64 + 64 * cos(w), y + 0 + 64 * sin(w));
    };
    float db = 20.0f * logf(sqrtf(radio_signal_strength_squared) / noise_level) / logf(10.0f);
    float angle = 150.0f - 120.0f * (db + 10) / 90.0f;
    if (angle > 150.0f) angle = 150.0f;
    if (angle < 30.0f) angle = 30.0f;
    float w = M_PI * angle / 180.0f;
    glColor3f(1.0, 0.2, 0.2);
    glVertex2f(x + 64, y + 0);
    glVertex2f(x + 64 + 58 * cos(w), y + 0 + 58 * sin(w));
    glEnd();
    smooth(0);
    glColor3f(1.0, 1.0, 1.0);
    gui_printf(event, x + 4, y + 0, "Signal-to-Noise");
    gui_printf(event, x + 28, y + 16, "%5.1f dB", db);
  };

  // Draw the demodulated signal strength meter...

  //x = window->width - 128; y = window->height - 72;
  x -= 128;
  if (!event) {
    smooth(1);
    if (!option_my_gpu_sucks) glLineWidth(1.5);
    glBegin(GL_LINES);
    for (float db = 0; db >= -80; db -= (option_my_gpu_sucks ? 5.0 : 0.5)) {
      float r, g, b;
      db_to_rgb(&r, &g, &b, db);
      glColor3f(r, g, b);
      float angle = 30.0f - 120.0f * db / 80.0f;
      if (angle > 150.0f) angle = 150.0f;
      float w = M_PI * angle / 180.0f;
      glVertex2f(x + 64 + 60 * cos(w), y + 0 + 60 * sin(w));
      glVertex2f(x + 64 + 64 * cos(w), y + 0 + 64 * sin(w));
    };
    glColor3f(1.0, 1.0, 1.0);
    for (int db = 0; db >= -80; db -= 10) {
      float angle = 30.0f - 120.0f * db / 80.0f;
      if (angle > 150.0f) angle = 150.0f;
      float w = M_PI * angle / 180.0f;
      glVertex2f(x + 64 + 60 * cos(w), y + 0 + 60 * sin(w));
      glVertex2f(x + 64 + 64 * cos(w), y + 0 + 64 * sin(w));
    };
    glColor3f(1.0, 0.2, 0.2);
    glVertex2f(x + 64, y + 0);
    float db = 20.0f * logf(sqrt(radio_signal_strength_squared)) / logf(10.0f);
    // full scale (0 db) = 60 degrees right
    float angle = 30.0f - 120.0f * db / 80.0f;
    if (angle > 150.0f) angle = 150.0f;
    float w = M_PI * angle / 180.0f;
    glVertex2f(x + 64 + 58 * cos(w), y + 0 + 58 * sin(w));
    glEnd();
    smooth(0);
    glColor3f(1.0, 1.0, 1.0);
    gui_printf(event, x + 4, y + 0, "Signal Strength");
    gui_printf(event, x + 32, y + 16, "%5.1f dB", db);
  };

  // Draw the carrier offset tuning meter...

  x -= 128;
  if (radio_state.mode == RADIO_MODE_AM_10 || radio_state.mode == RADIO_MODE_FM_12 || radio_state.mode == RADIO_MODE_FM_25) {
    y = window->height - 72;
    if (!event) {
      smooth(1);
      if (!option_my_gpu_sucks) glLineWidth(1.5);
      glBegin(GL_LINES);
      for (int i = -5; i <= 5; i++) {
        float deflection = i * 1000.0f / 30000.0f;
        float angle = 2 * M_PI * deflection;
        glVertex2f(x + 64 + 60 * sin(angle), y + 0 + 60 * cos(angle));
        glVertex2f(x + 64 + 64 * sin(angle), y + 0 + 64 * cos(angle));
      };
      glColor3f(1.0, 0.2, 0.2);
      glVertex2f(x + 64, y + 0);
      float deflection = radio_carrier_frequency_offset / 30000.0f;
      float limit = 1.0f / 6.0f;
      if (deflection < -limit) deflection = -limit;
      if (deflection > +limit) deflection = +limit;
      float angle = 2 * M_PI * deflection;
      glVertex2f(x + 64 + 58 * sin(angle), y + 0 + 58 * cos(angle));
      glEnd();
      smooth(0);
      glColor3f(1.0, 1.0, 1.0);
      gui_printf(event, x + 20, y + 16, "1 kHz / div");
      gui_printf(event, x + 8, y + 0, "Carrier Offset");
    };
  };

};


void user_interface () {

  memory_allocate(audio.buffer, 3 * audio.width * audio.height);
  MUTEX_INIT(audio.mutex);

  memory_allocate(signal.buffer, 3 * signal.width * signal.height);
  MUTEX_INIT(signal.mutex);

  memory_allocate(spectrum.buffer, 3 * spectrum.width * spectrum.height);
  MUTEX_INIT(spectrum.mutex);

  glfw_generic_event_loop(1024 + 256, 768 + 80, "Natpos SDR", 0, NULL, ui, NULL);

};

void user_initialize() {
};

void user_terminate() {
};
